import binascii
import json
from pymacaroons import utils


class JsonSerializer(object):
    '''Serializer used to produce JSON macaroon format v1.
    '''
    def serialize(self, m):
        '''Serialize the macaroon in JSON format indicated by the version field.

        @param macaroon the macaroon to serialize.
        @return JSON macaroon.
        '''
        from pymacaroons import macaroon
        if m.version == macaroon.MACAROON_V1:
            return self._serialize_v1(m)
        return self._serialize_v2(m)

    def _serialize_v1(self, macaroon):
        '''Serialize the macaroon in JSON format v1.

        @param macaroon the macaroon to serialize.
        @return JSON macaroon.
        '''
        serialized = {
            'identifier': utils.convert_to_string(macaroon.identifier),
            'signature': macaroon.signature,
        }
        if macaroon.location:
            serialized['location'] = macaroon.location
        if macaroon.caveats:
            serialized['caveats'] = [
                _caveat_v1_to_dict(caveat) for caveat in macaroon.caveats
            ]
        return json.dumps(serialized)

    def _serialize_v2(self, macaroon):
        '''Serialize the macaroon in JSON format v2.

        @param macaroon the macaroon to serialize.
        @return JSON macaroon in v2 format.
        '''
        serialized = {}
        _add_json_binary_field(macaroon.identifier_bytes, serialized, 'i')
        _add_json_binary_field(binascii.unhexlify(macaroon.signature_bytes),
                               serialized, 's')

        if macaroon.location:
            serialized['l'] = macaroon.location
        if macaroon.caveats:
            serialized['c'] = [
                _caveat_v2_to_dict(caveat) for caveat in macaroon.caveats
            ]
        return json.dumps(serialized)

    def deserialize(self, serialized):
        '''Deserialize a JSON macaroon depending on the format.

        @param serialized the macaroon in JSON format.
        @return the macaroon object.
        '''
        deserialized = json.loads(serialized)
        if deserialized.get('identifier') is None:
            return self._deserialize_v2(deserialized)
        else:
            return self._deserialize_v1(deserialized)

    def _deserialize_v1(self, deserialized):
        '''Deserialize a JSON macaroon in v1 format.

        @param serialized the macaroon in v1 JSON format.
        @return the macaroon object.
        '''
        from pymacaroons.macaroon import Macaroon, MACAROON_V1
        from pymacaroons.caveat import Caveat

        caveats = []
        for c in deserialized.get('caveats', []):
            caveat = Caveat(
                caveat_id=c['cid'],
                verification_key_id=(
                    utils.raw_b64decode(c['vid']) if c.get('vid')
                    else None
                ),
                location=(
                    c['cl'] if c.get('cl') else None
                ),
                version=MACAROON_V1
            )
            caveats.append(caveat)

        return Macaroon(
            location=deserialized.get('location'),
            identifier=deserialized['identifier'],
            caveats=caveats,
            signature=deserialized['signature'],
            version=MACAROON_V1
        )

    def _deserialize_v2(self, deserialized):
        '''Deserialize a JSON macaroon v2.

        @param serialized the macaroon in JSON format v2.
        @return the macaroon object.
        '''
        from pymacaroons.macaroon import Macaroon, MACAROON_V2
        from pymacaroons.caveat import Caveat
        caveats = []
        for c in deserialized.get('c', []):
            caveat = Caveat(
                caveat_id=_read_json_binary_field(c, 'i'),
                verification_key_id=_read_json_binary_field(c, 'v'),
                location=_read_json_binary_field(c, 'l'),
                version=MACAROON_V2
            )
            caveats.append(caveat)
        return Macaroon(
            location=_read_json_binary_field(deserialized, 'l'),
            identifier=_read_json_binary_field(deserialized, 'i'),
            caveats=caveats,
            signature=binascii.hexlify(
                _read_json_binary_field(deserialized, 's')),
            version=MACAROON_V2
        )


def _caveat_v1_to_dict(c):
    ''' Return a caveat as a dictionary for export as the JSON
    macaroon v1 format.
    '''
    serialized = {}
    if len(c.caveat_id) > 0:
        serialized['cid'] = c.caveat_id
    if c.verification_key_id:
        serialized['vid'] = utils.raw_urlsafe_b64encode(
            c.verification_key_id).decode('utf-8')
    if c.location:
        serialized['cl'] = c.location
    return serialized


def _caveat_v2_to_dict(c):
    ''' Return a caveat as a dictionary for export as the JSON
    macaroon v2 format.
    '''
    serialized = {}
    if len(c.caveat_id_bytes) > 0:
        _add_json_binary_field(c.caveat_id_bytes, serialized, 'i')
    if c.verification_key_id:
        _add_json_binary_field(c.verification_key_id, serialized, 'v')
    if c.location:
        serialized['l'] = c.location
    return serialized


def _add_json_binary_field(b, serialized, field):
    ''' Set the given field to the given val (a bytearray) in the serialized
    dictionary.

    If the value isn't valid utf-8, we base64 encode it and use field+"64"
    as the field name.
    '''
    try:
        val = b.decode("utf-8")
        serialized[field] = val
    except UnicodeDecodeError:
        val = utils.raw_urlsafe_b64encode(b).decode('utf-8')
        serialized[field + '64'] = val


def _read_json_binary_field(deserialized, field):
    ''' Read the value of a JSON field that may be string or base64-encoded.
    '''
    val = deserialized.get(field)
    if val is not None:
        return utils.convert_to_bytes(val)
    val = deserialized.get(field + '64')
    if val is None:
        return None
    return utils.raw_urlsafe_b64decode(val)



from .first_party import (
    FirstPartyCaveatDelegate, FirstPartyCaveatVerifierDelegate
)
from .encrypted_first_party import (
    EncryptedFirstPartyCaveatDelegate,
    EncryptedFirstPartyCaveatVerifierDelegate
)
from .third_party import (
    ThirdPartyCaveatDelegate,
    ThirdPartyCaveatVerifierDelegate
)

__all__ = [
    'FirstPartyCaveatDelegate',
    'FirstPartyCaveatVerifierDelegate',
    'ThirdPartyCaveatDelegate',
    'ThirdPartyCaveatVerifierDelegate',
    'EncryptedFirstPartyCaveatDelegate',
    'EncryptedFirstPartyCaveatVerifierDelegate',
]



from abc import ABCMeta, abstractmethod


class BaseBinder(object):
    __metaclass__ = ABCMeta

    def __init__(self, root):
        self.root = root

    def bind(self, discharge):
        protected = discharge.copy()
        protected.signature = self.bind_signature(discharge.signature_bytes)
        return protected

    @abstractmethod
    def bind_signature(self, signature):
        pass



import binascii

from pymacaroons import Caveat
from pymacaroons.utils import (
    convert_to_string,
    convert_to_bytes,
    sign_first_party_caveat
)

from .base_first_party import (
    BaseFirstPartyCaveatDelegate,
    BaseFirstPartyCaveatVerifierDelegate
)


class FirstPartyCaveatDelegate(BaseFirstPartyCaveatDelegate):

    def __init__(self, *args, **kwargs):
        super(FirstPartyCaveatDelegate, self).__init__(*args, **kwargs)

    def add_first_party_caveat(self, macaroon, predicate, **kwargs):
        predicate = convert_to_bytes(predicate)
        # Check it's valid utf-8 for first party caveats.
        # Will raise a unicode error if not.
        predicate.decode('utf-8')
        caveat = Caveat(caveat_id=predicate, version=macaroon.version)
        macaroon.caveats.append(caveat)
        encode_key = binascii.unhexlify(macaroon.signature_bytes)
        macaroon.signature = sign_first_party_caveat(encode_key, predicate)
        return macaroon


class FirstPartyCaveatVerifierDelegate(BaseFirstPartyCaveatVerifierDelegate):

    def __init__(self, *args, **kwargs):
        super(FirstPartyCaveatVerifierDelegate, self).__init__(*args, **kwargs)

    def verify_first_party_caveat(self, verifier, caveat, signature):
        predicate = convert_to_string(caveat.caveat_id)
        caveat_met = sum(callback(predicate)
                         for callback in verifier.callbacks)
        return caveat_met

    def update_signature(self, signature, caveat):
        return binascii.unhexlify(
            sign_first_party_caveat(
                signature,
                caveat.caveat_id_bytes
            )
        )



import binascii

from pymacaroons.field_encryptors import SecretBoxEncryptor

from .first_party import (
    FirstPartyCaveatDelegate, FirstPartyCaveatVerifierDelegate
)


class EncryptedFirstPartyCaveatDelegate(FirstPartyCaveatDelegate):

    def __init__(self, field_encryptor=None, *args, **kwargs):
        self.field_encryptor = field_encryptor or SecretBoxEncryptor()
        super(EncryptedFirstPartyCaveatDelegate, self).__init__(
            *args, **kwargs
        )

    def add_first_party_caveat(self, macaroon, predicate, **kwargs):
        if kwargs.get('encrypted'):
            predicate = self.field_encryptor.encrypt(
                binascii.unhexlify(macaroon.signature_bytes),
                predicate
            )
        return super(EncryptedFirstPartyCaveatDelegate,
                     self).add_first_party_caveat(macaroon,
                                                  predicate,
                                                  **kwargs)


class EncryptedFirstPartyCaveatVerifierDelegate(
        FirstPartyCaveatVerifierDelegate):

    def __init__(self, field_encryptors=None, *args, **kwargs):
        secret_box_encryptor = SecretBoxEncryptor()
        self.field_encryptors = dict(
            (f.signifier, f) for f in field_encryptors
        ) if field_encryptors else {
            secret_box_encryptor.signifier: secret_box_encryptor
        }
        super(EncryptedFirstPartyCaveatVerifierDelegate, self).__init__(
            *args, **kwargs
        )

    def verify_first_party_caveat(self, verifier, caveat, signature):
        predicate = caveat.caveat_id_bytes

        for signifier, encryptor in self.field_encryptors.items():
            if predicate.startswith(signifier):
                predicate = encryptor.decrypt(
                    signature,
                    predicate
                )

        caveat_met = sum(callback(predicate)
                         for callback in verifier.callbacks)
        return caveat_met
